/*////////////////////////////////////////////////////////////////////////
Copyright (c) 2005-2006 National Institute of Advanced Industrial Science and Technology (AIST)

Permission to use this material for noncommercial and/or evaluation
purpose, copy this material for your own use, and distribute the copies
via publicly accessible on-line media, without fee, is hereby granted
provided that the above copyright notice and this permission notice
appear in all copies.
AIST MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	gzip.c
Author:		Yutaka Sato <ysato@delegate.org>
Description:

History:
	050501	created
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>
#include <stdlib.h>
#include "ystring.h"
#include "log.h"
int daemonlog(PCStr(what),PCStr(fmt),...);

/*BEGIN_STAB(zlib)*/
typedef void *gzFile;
const char *zlibVersion();
gzFile gzopen(const char *path,const char *mode);
gzFile gzdopen(int fd,const char *mode);
int gzwrite(gzFile file,const void *buf,unsigned len);
int gzread(gzFile file,void *buf,unsigned len);
int gzclose(gzFile file);
const char *gzerror(gzFile file,int *errnum);
typedef unsigned char Byte;
typedef unsigned long uLong;
int compress2(Byte *dest,uLong *destLen,const Byte *source,uLong sourceLen,int lev);
int uncompress(Byte *dest,uLong *destLen,const Byte *source,uLong sourceLen);
/*END_STAB*/

int dl_library(const char *libname,DLMap *dlmap,const char *mode);
#ifdef _MSC_VER
int gzipInit0(){
	int code;
	code = dl_library("dgzlib1",dlmap_zlib,"");
	return code;
}
int fd2handle(int fd);
gzFile GZdopen(int fd,const char *mode){
	gzFile gz;
	int handle;

	handle = 0x80000000 | fd2handle(fd);
	gz = gzdopen(handle,mode);
	if( gz == 0 ){
		syslog_ERROR("-- failed gzdopen(0x%X)\n",handle);
		gz = gzdopen(fd,mode);
	}
	if( gz == 0 ){
		syslog_ERROR("-- failed gzdopen(%d)\n",fd);
	}
	return gz;
}
#else
int gzipInit0(){
	int code;
	code = dl_library("z",dlmap_zlib,"");
	return code;
}
#define GZdopen(fd,mode) gzdopen(fd,mode)
#endif

static int zlib_dl;
int gzipInit(){
	int code;
	code = gzipInit0();
	if( code == 0 ){
		InitLog("+++ loaded Zlib %s\n",zlibVersion());
		if( lDYLIB() )
		printf("+++ loaded Zlib %s\n",zlibVersion());
		zlib_dl = 1;
	}else{
		zlib_dl = -1;
	}
	return code;
}
void putZLIBver(FILE *fp){
	if( 0 < zlib_dl )
		fprintf(fp,"Loaded: Zlib %s\r\n",zlibVersion());
}

#include "file.h"

#ifdef MMAP
#ifdef _MSC_VER /*{*/
#define PROT_READ 1
#define PROT_WRITE 2
#define MAP_SHARED 0
static HANDLE last_fmh[2];
void *mmap(void *adr,size_t len,int pro,int flg,int fd,off_t off){
	void *addr;
	HANDLE fh,fmh;
	int protect;
	int acc;
	int rw;

	fh = (HANDLE)_get_osfhandle(fd);
	if( pro == PROT_READ ){
		protect = PAGE_READONLY;
		acc = FILE_MAP_READ;
	}else{
		protect = PAGE_READWRITE;
		acc = FILE_MAP_WRITE;
	}

	fmh = CreateFileMapping(fh,NULL,protect,0,off+len,NULL);
	if( last_fmh[0] == 0 )
		last_fmh[0] = fmh;
	else	last_fmh[1] = fmh;
	addr = MapViewOfFile(fmh,acc,0,off,len);
	return addr;
}
int munmap(void *adr,size_t le){
	int ok = UnmapViewOfFile(adr);
	if( last_fmh[0] ){
		CloseHandle(last_fmh[0]);
		last_fmh[0] = 0;
	}
	if( last_fmh[1] ){
		CloseHandle(last_fmh[1]);
		last_fmh[0] = 0;
	}
	return ok ? 0 : -1;
}

#else /*}{*/
#include <sys/mman.h>
#endif /*}*/

int gzipMmap(int do_comp,FILE *in,FILE *out){
	double Start = Time();
	int ifd,ofd;
	int iz,izm;
	unsigned long oz;
	Byte *ia,*iam;
	Byte *oa,*oam;
	int rcode;
	int ioff;
	int ooff;
	int ilen = 0;

	ifd = fileno(in);
	ofd = fileno(out);
	if( !file_isreg(ifd) || !file_isreg(ofd) ){
		syslog_ERROR("--- gzipMmap: not reg-file: %d %d\n",ifd,ofd);
		return -1;
	}

	ioff = lseek(ifd,0,1);
	ooff = lseek(ofd,0,1);
	izm = file_size(ifd);
	iz = izm - ioff;
	if( do_comp )
		oz = 1024+iz;
	else	oz = 1024+iz*20;

	iam = (Byte*)mmap(0,izm,PROT_READ,MAP_SHARED,ifd,0);
	if( iam == 0 ){
		syslog_ERROR("--- gzipMmap: can't open in mmap(%d)\n",ifd);
		return -1;
	}
	ia = iam + ioff;

	oa = (Byte*)mmap(0,oz,PROT_READ|PROT_WRITE,MAP_SHARED,ofd,ooff);
	if( oa == 0 ){
		syslog_ERROR("--- gzipMmap: can't open out mmap(%d)\n",ofd);
		munmap(iam,iz);
		return -1;
	}

	lseek(ofd,oz-1,1);
	write(ofd,"",1);
	if( do_comp ){
		/*
		Byte *op = oa;
		*op++ = 0x1F; *op++ = 0x8B; *op++ = 8;
		*op++ = 0; *op++ = 0; *op++ = 0; *op++ = 0;
		*op++ = 0; *op++ = 0; *op++ = 3;
		rcode = compress2(op,&oz,ia,iz,-1);
		if( rcode == 0 ){
			oz += (op - oa);
		}
		*/
		rcode = compress2(oa,&oz,ia,iz,-1);
	}else{
		rcode = uncompress(oa,&oz,ia,iz);
	}

	munmap(iam,izm);
	munmap(oa,oz);

	syslog_ERROR("(%.4f)g%szip/mmap(%d) %d -> %d\n",
		Time()-Start,do_comp?"":"un",rcode,iz,oz);

	if( rcode == 0 ){
		Ftruncate(out,ooff+oz,0);
		fseek(out,0,0);
		if( do_comp )
			return iz;
		else	return oz;
	}
	lseek(ifd,ioff,0);
	lseek(ofd,ooff,0);
	Ftruncate(out,0,1);
	return -1;
}
#endif

int gzipFilter(FILE *in,FILE *out){
	gzFile gz;
	int len,rcc;
	char buf[1024*8];
	/*
	char buf[1024];
	*/
	int size;
	double Start = Time();

	len = 0;
	/*
	if( gz = gzdopen(dup(fileno(out)),"w") ){
	*/
	if( gz = GZdopen(dup(fileno(out)),"w") ){
		while( rcc = fread(buf,1,sizeof(buf),in) ){
			gzwrite(gz,buf,rcc);
			len += rcc;
		}
		gzclose(gz);
		fseek(out,0,2);
		size = ftell(out);
		fseek(out,0,0);
		syslog_DEBUG("(%f)gzipFilter %d -> %d\n",Time()-Start,len,size);
		return len;
	}
	return 0;
}
int gunzipFilter(FILE *in,FILE *out){
	gzFile gz;
	int rcc;
	char buf[1024*8];
	/*
	char buf[1024];
	*/
	int size;
	double Start = Time();
	const char *em;
	int en;

	size = 0;
	/*
	if( gz = gzdopen(dup(fileno(in)),"r") ){
	*/
	if( gz = GZdopen(dup(fileno(in)),"r") ){
		em = gzerror(gz,&en);
		/*
		while( rcc = gzread(gz,buf,sizeof(buf)) ){
		*/
		while( 0 < (rcc = gzread(gz,buf,sizeof(buf))) ){
			fwrite(buf,1,rcc,out);
			size += rcc;
		}
		if( rcc < 0 ){
			em = gzerror(gz,&en);
			daemonlog("F","FATAL: gzread()=%d %d %s\n",rcc,en,em);
		}
		gzclose(gz);
		fseek(out,0,0);
		syslog_DEBUG("(%f)gunzipFilter -> %d\n",Time()-Start,size);
		return size;
	}
	return 0;
}
int zlibUncompress(void *in,int isiz,void *out,int osiz){
	int stat;
	int dlen;

	dlen = osiz;
	stat = uncompress((Byte*)out,(uLong*)&dlen,(const Byte*)in,(uLong)isiz);
	if( stat == 0 )
		return dlen;
	else	return -1;
}
